=begin

  *Name: Atlas_SNP_SAM_v1.0.rb 
  *Description:
      'Atlas_SNP_SAM_v1.0.rb' is a SNP caller which takes alignments in SAM format and reference sequences as inputs to carry out putative SNP calling and posterior probabilities evaluation. The current release can work on data from 454 FLX and XLR.
  *Authors: Jin Yu, Yong (Tony) Wang and Fuli Yu @ BCM - Human Genome Sequencing Center
  *Date:  07/29/2009

=end



require 'getoptlong'
require 'bigdecimal'

#start of class definition

class AtlasSNP_SAM

#Logistic regression coefficients
$logistic_FLX={:intercept=>-5.36, :quality_score=>0.11, :swap=>-4.88, :NQS_pass=>1.07, :relpos=>0.91}
$logistic_XLR={:intercept=>-3.31, :quality_score=>0.1, :NQS_pass=>0.26, :swap=>-3.5, :relpos=>-0.37}

#Sj (variant read coverage <= 2)initialization
$errorPrior_test_less_3=[0.867, 0.867, 0.062, 0.062, 0.063, 0.063, 0.008, 0.008, 0.0003, 0.0003]
$snpPrior_test_less_3=[0.78, 0.78, 0.122, 0.122, 0.049, 0.049, 0.049, 0.049, 0.00001, 0.00001]

#Sj (variant read coverage >= 3)initialization
$errorPrior_test_greater_3=[0.676, 0.676, 0.102, 0.102, 0.074, 0.074, 0.028, 0.028, 0.12, 0.12]
$snpPrior_test_greater_3=[0.00001, 0.00001, 0.006, 0.006, 0.006, 0.006, 0.041, 0.041, 0.947, 0.947]


$basechange={"A"=>"T","T"=>"A","G"=>"C","C"=>"G","N"=>"N","X"=>"X","*"=>"*","a"=>"t","t"=>"a","g"=>"c","c"=>"g"}


def AtlasSNP_SAM.processArguments()

	opts = GetoptLong.new(
    		["--input", "-i", GetoptLong::REQUIRED_ARGUMENT],
    		["--reference", "-r", GetoptLong::REQUIRED_ARGUMENT],
    		["--help","-h", GetoptLong::NO_ARGUMENT],
    		["--output","-o", GetoptLong::REQUIRED_ARGUMENT],
    		["--maxsub", "-s", GetoptLong::OPTIONAL_ARGUMENT],
    		["--maxindel", "-g", GetoptLong::OPTIONAL_ARGUMENT],
    		["--XLR", "-x", GetoptLong::NO_ARGUMENT],
    		["--priorerror","-e", GetoptLong::OPTIONAL_ARGUMENT]
	)

	optHash = {}
	opts.each do |opt, arg|
  		optHash[opt] = arg
	end

	AtlasSNP_SAM.usage() if (optHash.key?("--help"));
	AtlasSNP_SAM.usage() if (optHash.empty?);
	return optHash
end

def AtlasSNP_SAM.usage(msg='')
	unless (msg.empty?)
        	puts "\n#{msg}\n"
    	end
    		puts "  
'Atlas_SNP_SAM_v1.0.rb' is a SNP caller which takes alignments in SAM format and reference sequences as inputs to carry out putative SNP calling and posterior probabilities evaluation. The current release can work on data from 454 FLX and XLR.

Required arguments:

-i input SAM file 
-r Reference sequence file
-o output file name (SNP list with evaluations)

Optional arguments:

-s the maximum number of substitutions contained in a variant read to be a true variant read.
-g the maximum number of indels contained in a variant read to be a true variant read.
-x This select must be appended if the input is XLR data.
-e with an associate value, Prior(error|c) when variant coverage greater than 2. The default value is 0.1 if this option is ignored.
 
Output Format:

refName\tcoordinate\trefBase\thomopolymer\trefEnv\tvariantBase\tvarCoverage\tAlternativereads\ttotalQual\treadsInfo\tPr(error)j\tPr(SNP)j\tPr(Sj|error,c)\tPr(Sj|SNP,c)\tPrior(error|c)\tPrior(SNP|c)\tPosterior(SNP|Sj,c)j

readinfo: base(qual)name(dist)(score/len)dir env(sub/gap/tail)(nqs/len)swap


"
	exit(2);
end

def initialize(optHash)
	@optHash=optHash
	setParameters()
end	

def setParameters()

	if (!@optHash.key?('--input') or !@optHash.key?("--reference") or !@optHash.key?("--output") )then
        	AtlasSNP_SAM.usage("Option missing!")
        	exit(2);
    	end
    
	#set input files

	@SAMFile=@optHash["--input"]
	@refFile=@optHash["--reference"]

	#set parameters

	if @optHash.key?("--maxsub")
		@maxsub = @optHash["--maxsub"].to_f
	else
  		@maxsub = 5.0
	end

	if @optHash.key?("--maxindel")
		@maxindel = @optHash["--maxindel"].to_f
	else	
		@maxindel = 5.0
	end
	
	if @optHash.key?("--XLR")
		@XLR=true
		@logistic=$logistic_XLR
	else
		@XLR=false
		@logistic=$logistic_FLX
	end

	if @optHash.key?("--priorerror")
		@perror_highcov=@optHash["--priorerror"]
  		@psnp_highcov=1-@perror_highcov.to_f
	else
		@perror_highcov=0.1
		@psnp_highcov=0.9
	end

	#set output files
	@outputFile=@optHash["--output"]
	@missedFile=@optHash["--output"]+".missed_reads"
end


#Hash refer sequence to @seq[ref]="str" , prepare a corresponding @snp[ref]
def hashReference()
    	ref = ''
	@seq={}
	@snp={}
    	@coverage={}
	refReader=File.open(@refFile,"r")
	refReader.each do |line|
		if line=~ /^>(\S+)/
			ref = $1
			if line=~ /^>chr(\S+)/
				ref = $1
			end	
          		@seq[ref] = ''
          		@snp[ref] = {}
                	@coverage[ref] = {}
        	else
          		@seq[ref] << line.chomp
        	end
    	end
    	refReader.close
end


def SNP_list()
	snps = {}
     
	sam_Reader=File.open(@SAMFile,"r")
    
	sam_Reader.each do |line|

	        next if line.match(/^@/)

       		read_sam_cols = line.split
        
        	query = read_sam_cols[0]
        	dir = read_sam_cols[1]
        	ref = read_sam_cols[2]
        	query_coord = read_sam_cols[3].to_i
        	cigar = read_sam_cols[5]
        	read_seq = read_sam_cols[9]
		qual = read_sam_cols[10]

		num_fields = read_sam_cols.length
		score = ""
		md = ""
		sub = ""
		del = ""
		ins = ""
		(11..(num_fields-1)).each do |col_num|
			if read_sam_cols[col_num] =~ /^AS:i:(\S+)$/
				score = $1
				next
			end
			if read_sam_cols[col_num] =~ /^XP:f:(\S+)$/
                		sub = $1
                		next
            		end
			if read_sam_cols[col_num] =~ /^XD:f:(\S+)$/
                		del = $1
                                next
                        end
			if read_sam_cols[col_num] =~ /^XI:f:(\S+)$/
                                ins = $1
                                next
                        end
			if read_sam_cols[col_num] =~ /^MD:Z:(\S+)$/
                                md = $1
                                next
                        end
		end

		# get information about this query which are general to all the potential SNP detected in this read query

        	gap = del.to_f + ins.to_f
        
		# filter these queries
        
        	if (sub.to_f > @maxsub) or (gap > @maxindel)
                	next
        	end

        	#computer total coverage
        
            	span = Parse_CIGAR_query_span(cigar)
            	all_del_span = Parse_CIGAR_del_span(cigar,dir)
            	Calculate_total_coverage(cigar, query_coord, span, all_del_span, ref)

        	query_size = read_seq.length
        	qual_size = qual.length
            
           
        	softclip = Parse_CIGAR_softclip(cigar).to_i
            	softclip_end = Parse_CIGAR_softclip_end(cigar).to_i
            
            	if dir == "16"
                	tail = [softclip_end + 1, softclip].max
            	else
                	tail = [softclip_end, softclip + 1].max
            	end
            
            

		# retrieve information of all potential SNPs detected in this read query

		# A potential SNP can be completely located by 2 different coordinates (query, query_place) and 
		# (ref, tplace), information needed include: dist, tstart, NQS, varbase, query_env, refer_env, SNP/SWAP/NMP

		#parse MD code to get qplace of all potential SNP in this query, generate a snp list of this query by using these information.

        	md_parsed = md.sub(/(\d+)$/,"")
        	md_parsed.sub!(/(\d+)$/,"")
        	md_parsed.gsub!(/(\^[TCAGatgc]+)/,"/")
        	md_parsed.gsub!(/[TCAGatgc]/,"/flag/")
        
        	snp_pos_list = md_parsed.split("/")

        	snp_list = {}

		# use this hash table to store all the SNP in the index of reference coordinate, also for the purpose of reusing original codes as much as possible
        	@snps = {} 

        	i = 0
        	num = 0
        	place = 0

        	snp_pos_list.each do |part|
        
        		if part == "flag" 	#use "flag" to detect a SNP
               			place += 1
                		snp_list[num] = {}
                		snp_list[num][:query] = query

                		snp_list[num][:query_place] = place + softclip
                
                		snp_list[num][:dist] = query_size - snp_list[num][:query_place]
                		snp_list[num][:qual] = qual[snp_list[num][:query_place] - 1 ].to_i - 33
                		snp_list[num][:tplace] = Parse_CIGAR_snp_tplace(cigar, query_coord, snp_list[num][:query_place]) -softclip


                		snp_list[num][:ii] = read_seq[(snp_list[num][:query_place] - 1), 1]

                		#format readEnv of this SNP
                		if snp_list[num][:query_place] >= 7
                    			snp_list[num][:readenv] = read_seq[(snp_list[num][:query_place] - 7), 6].downcase + read_seq[(snp_list[num][:query_place] - 1), 1].upcase + read_seq[snp_list[num][:query_place]..(snp_list[num][:query_place] + 5)].downcase
                		else
                    			snp_list[num][:readenv] = read_seq[0..(snp_list[num][:query_place]-2)].downcase + read_seq[(snp_list[num][:query_place] - 1), 1].upcase + read_seq[snp_list[num][:query_place]..(snp_list[num][:query_place] + 5)].downcase
                		end
                
                    # make modifications to dist when detect a reversed read
                		if dir == "16"                            
                			snp_list[num][:dist] = query_size - snp_list[num][:dist] - 1
                		end
                        
                		snp_list[num][:nqspass] = Calculate_NQSpass(snp_list[num][:dist], qual, qual_size, dir)

                		#add this potential snp to @snps, ignore "N" ones
                		next if snp_list[num][:ii] == "N" 
                
                		@snps[snp_list[num][:tplace]] = {}
                		@snps[snp_list[num][:tplace]][:snpbase] = snp_list[num][:ii]
                		@snps[snp_list[num][:tplace]][:info] = "#{snp_list[num][:readenv]}\t#{snp_list[num][:ii]}\t#{snp_list[num][:qual]}\t#{snp_list[num][:query]}\t#{snp_list[num][:query_place]}\t#{snp_list[num][:dist]}\t#{qual_size}\t#{snp_list[num][:nqspass]}\t#{score}\t#{sub}\t#{gap}\t#{ins}\t#{tail}\t#{cigar}\t#{dir}\t"
                
                		num += 1
			end 

		    	place += part.to_i 
        end 	
        
    #ends calculation of all SNPs in this query
    
		#compute SNP/SWAP/NMP and hash SNPs in thie query to @snp[ref][pos]       
        	compute_swap(ref,@snps) 
	end
	sam_Reader.close
end

def compute_swap(ref,snps)
	snps.each_key do |pos|
		pos = pos.to_i
		refbase = @seq[ref][pos-1,1].upcase
		curbase = snps[pos][:snpbase]
		if snps.key?(pos + 1) or snps.key?(pos + 2) or snps.key?(pos - 1) or snps.key?(pos - 2)
			if snps.key?(pos + 1) and refbase == snps[pos+1][:snpbase] and curbase == @seq[ref][pos,1].upcase
				snps[pos][:info] << "swap;"
			elsif snps.key?(pos - 1) and refbase == snps[pos-1][:snpbase] and curbase == @seq[ref][pos-2,1].upcase 
				snps[pos][:info] << "swap;"
			elsif snps.key?(pos + 2) and refbase == snps[pos+2][:snpbase] and curbase == @seq[ref][pos+1,1].upcase
				snps[pos][:info] << "swap;"
			elsif snps.key?(pos - 2) and refbase == snps[pos-2][:snpbase] and curbase == @seq[ref][pos-3,1].upcase
				snps[pos][:info] << "swap;"
			elsif snps.key?(pos + 1) or snps.key?(pos - 1) 
				snps[pos][:info] << "mnp;"
			else
				snps[pos][:info] << "snp;"        
			end
		else
			snps[pos][:info] << "snp;"      
		end
		@snp[ref][pos] = '' unless @snp[ref].key?(pos)
		@snp[ref][pos] << snps[pos][:info] 
	end
end  

def evaluate()
	@errPrior=[] 
	@snpPrior=[]
	prior_error_c=-10
	prior_snp_c=-10
	var_cov=0

	outputWriter=File.new(@outputFile,"w")
	missedWriter=File.new(@missedFile,"w")
	outputWriter.print "refName\tcoordinate\trefBase\thomopolymer\trefEnv\ttotalCoverage\tvariantBase\toriQual\tvariantReadCov\tnumAlternativeReads\treadsInfo\tPr(error)j\tPr(SNP)j\tPr(Sj|error,c)\tPr(Sj|SNP,c)\tPrior(error|c)\tPrior(SNP|c)\tPosterior(SNP|Sj,c)j\n"
 
	@snp.keys.sort.each do |ref|
		@snp[ref].keys.sort.each do |pos|   
        		base_num = {}                                               #prepare it for calculating major varbase
                	readinfo_output = '' 
        		added_qual = 0
        		refbase=@seq[ref][pos.to_i - 1, 1]
        		refenv=@seq[ref][pos.to_i - 7, 13]
                	totalcoverage=@coverage[ref][pos.to_i].to_i
                	alternativereads=@snp[ref][pos].split(';').length
                	homonum = homocount(refenv)

			# determine major varbase

        		@snp[ref][pos].split(';').each do |r|   
        			col = r.split("\t")
            			base=col[1]
            			if base_num[base] == nil
                			base_num[base] = {}
                            		base_num[base][:num] = 1
            			else
                			base_num[base][:num] += 1
            			end
        		end
        		array = base_num.keys.sort {|a,b| base_num[b][:num] <=> base_num[a][:num]}
                	varbase = array[0]
        		var_cov=0                                                          #reset coverage number
        		rstr=''                                                               #reset rset
        		pr_error_j=1                                                      #reset pr_error_j


    			@snp[ref][pos].split(';').each do |r|                   #start reading SNP list in this query 
            			read_col = r.split("\t")
                        	env  = read_col[0]
            			base = read_col[1]
            			qual = read_col[2].to_i
            			name = read_col[3]
                        	score= read_col[8].to_f
            			sub  = read_col[9]
                        	gap  = read_col[10]
                        	tail = read_col[12]
            			nqs  = read_col[7].to_i
            			dist = read_col[5].to_i
            			len  = read_col[6].to_f
                        	dir  = read_col[14]
            			swap = read_col[15]
            			relpos=dist/len.to_f
                
                		#format readinfo output     
                
                		if dir == "16"
                    			dir = "-"
                    			env.reverse!
                    			env.gsub!(/[atcgATCGNX*]/) {|s| s[0]=$basechange[s[0,1]]}
                		else
                    			dir = "+"
                		end
                
                		readinfo_output = "#{base}(#{qual})#{name}(#{dist})(#{score}/#{len.to_i})#{dir}#{env}(#{sub}/#{gap}/#{tail})(#{nqs}/#{len.to_i})#{swap}"

				next if base != varbase                                             #ignore SNP which are not major
				next if nqs==nil or len==nil                                     #prevent error

        			added_qual += qual                                        
        
        			if base == 'N'                                                     #This should not happen, as "N" ones have been ignored before
          				logOdd = 0
          				errorPosteriorFull = 1
        			else
          				if swap =="swap"                                           #set swap value
            					swap = 1
          				elsif swap =="mnp"
            					swap = 0.5
          				else
            					swap = 0
          				end
          
          				#Logistic regression p-value
          				predict = @logistic[:intercept] + @logistic[:quality_score]*qual.to_f + @logistic[:swap]*swap.to_f + @logistic[:NQS_pass]*nqs.to_f + @logistic[:relpos] * relpos.to_f
          				logisticP = Math.exp(predict) / (1 + Math.exp(predict))
          
          				pr_error_i=1-logisticP
          				pr_error_j*=pr_error_i
        			end
        
        			var_cov+=1                                                                  #var covaerage number adds 1
        
        			rstr<<readinfo_output+"("+pr_error_i.to_s+");"        #save pr_error_i of this query in rstr
			end                                                                                       #end process this SNP
  
			#all SNPs in this position have been processed,  pr_error_j is prepared
  
			if var_cov==0                                                                       #what if not found SNP in this position, This should not happen 
         			missedWriter.print "#{@snp[ref][pos]}\n"
    			end
  
			#Select Sj
			if var_cov>0
				outputWriter.print  "chr#{ref}\t#{pos}\t#{refbase}\t#{homonum}\t#{refenv}\t#{totalcoverage}\t#{varbase}\t#{added_qual}\t#{var_cov}\t#{alternativereads}\t#{rstr}\t"                                                     #Print out these information first
    

				pr_snp_j=1-BigDecimal(pr_error_j.to_s)
				bin = (BigDecimal(pr_snp_j.to_s) *10).floor                       #Determine the bin of Sa
  
				if bin==10
					bin=9
				end
    
				#Determine Sj sets according to variant reads coverage.
				if var_cov<3
					@snpPrior=$snpPrior_test_less_3 #Select Sj sets for SNPs
      					@errPrior=$errorPrior_test_less_3
      					prior_error_c=0.9 #Prior(Error|Cov)
      					prior_snp_c=0.1 #Prior(SNP|Cov)
    				elsif var_cov>=3
      					@errPrior=$errorPrior_test_greater_3
      					@snpPrior=$snpPrior_test_greater_3
      					prior_error_c=@perror_highcov.to_f
      					prior_snp_c=@psnp_highcov.to_f
    				end
  
  				#Get Pr(Sj|error,c) and Pr(Sj|SNP,c)
  				pr_Sj_error_c = @errPrior[bin]
  				pr_Sj_SNP_c = @snpPrior[bin]
  
  				#For debugging
  				if pr_Sj_error_c==nil or pr_Sj_SNP_c==nil
    					$stderr.puts bin
    					$stderr.puts prior_error_c
    					$stderr.puts prior_snp_c
    					$stderr.puts @snp[ref][pos]
    				end

  				errorPosterior = pr_Sj_error_c * prior_error_c
  				snpPosterior = pr_Sj_SNP_c * prior_snp_c
  
  				#Get Posterior(SNP|Sj,Cov)j
  				pr_SNP_Sj_c_j = 1.0 / (1 + errorPosterior/snpPosterior)
  
  				#print the useful variables and final Posterior(SNP|Sj,Cov)j
  				outputWriter.print "#{pr_error_j}\t#{pr_snp_j}\t#{pr_Sj_error_c}\t#{pr_Sj_SNP_c}\t#{prior_error_c}\t#{prior_snp_c}\t#{pr_SNP_Sj_c_j}\n"
			end
		end
	outputWriter.close
	missedWriter.close
	end
end


def Parse_CIGAR_softclip(cigar)
	cigar.match(/(^[\d]+)S/)
	return $1.to_i
end

def Parse_CIGAR_softclip_end(cigar)
	cigar.match(/([\d]+)S$/)
	return $1.to_i
end

#calculate coverage span(on reference) of this query, only consider M and D at here

def Parse_CIGAR_query_span(cigar)
	count = 0
	list = cigar.scan(/([\d]+[MD])/)
    
	list.each do |part|
		part.to_s.match(/([\d]+)/)
        count += $1.to_i
   	end
    
    	return count
end

#calculate the del span coverage (on reference) in this query

def Parse_CIGAR_del_span(cigar,dir)
    	count = 0
    	all_del_span = []
    	list = cigar.scan(/([\d]+[MD])/)
    
	list.each do |part|
		part.to_s.match(/([\d]+)([MD])/)
        
        	if $2 == 'M'
            		count += $1.to_i
        	elsif $2 == 'D'
            		one_del_span = []
            		one_del_span[0] = count                     #count before this del span           
            		one_del_span[1] = $1.to_i                   #length of this del span
            		all_del_span << one_del_span
            		count += $1.to_i                                  #count reference position
        	end
   	end
    	return all_del_span
end

#a special feature in ruby: the data type associated with an empty key in a hash table is nil, nil can be converted to 0

def Calculate_total_coverage(cigar, offset, query_span, all_del_span, ref)
    	(offset..(offset + query_span - 1)).each do |i|
        	@coverage[ref][i] =  @coverage[ref][i].to_i + 1    
    	end
    
    	while all_del_span.size > 0 do
       		one_del_span = all_del_span.shift
       		del_start = one_del_span[0].to_i
       		del_span  = one_del_span[1].to_i
       		((offset + del_start)..(offset + del_start + del_span - 1)).each do |i|
            		@coverage[ref][i] =  @coverage[ref][i].to_i - 1
        	end
    	end
end
  


def Parse_CIGAR_snp_tplace(cigar, coord, qplace)
    	count = 0
	num_D = 0
	num_I = 0
	num_S = 0
	tplace = 0
	list = cigar.scan(/([\d]+[MISD])/)

	list.each do |part|
        	part.to_s.match(/([\d]+)([MIDS])/)
        
        	count += ($1).to_i
        
        	if $2 == "I"
            		num_I += ($1).to_i
        	elsif $2 == "D"
            		num_D += ($1).to_i
            		count -= ($1).to_i
        	end
        
        	if count >= qplace
            		tplace = coord + qplace - num_I + num_D - 1
            		return tplace
        	end
    	end
end


    
def Calculate_NQSpass(dist, qual, count, dir)
	if dir =='16'
        	qual = qual.reverse
    	end
    
    	pos = count - 1 - dist
	if count > 10
		if dist <5
			p_status= 0
		elsif pos<5
			p_status= 1
		else
			pos_center= (qual[pos].to_i - 33)
			if qual[pos].to_i>=53 and qual[pos-1].to_i>=48 and qual[pos-2].to_i>=48 and qual[pos-3].to_i>=48 and qual[pos-4].to_i>=48 and qual[pos-5].to_i>=48 and qual[pos+1].to_i>=48 and qual[pos+2].to_i>=48 and qual[pos+3].to_i>=48 and qual[pos+4].to_i>=48 and qual[pos+5].to_i>=48
				p_status= 1
			else
                 		p_status= 0
             		end
         	end
    	else
        	p_status= 1
    	end
    	return p_status
end




def homocount(str)
  	s = str.size
  	lasts = ''
  	homo = {}
  	st = 0
  	flag = 0
  	0.upto(s-1) do |i|
    		cur = str[i,1].upcase
    		if cur == lasts
      			if flag == 0 # this is the second 
        			homo[i-1] = 2
        			st = i - 1
      			else # this is the third or more
        			homo[st] += 1
      			end
      			flag = 1
    		else
      			flag = 0
    		end
    		lasts = cur
  	end
  	if homo.size > 0
    		top = homo.keys.sort {|a,b| homo[b] <=> homo[a]}[0]
    		xx = homo[top]
  	else
    		xx = 1
  	end
  	return xx
end

end

# end of class definition


#Main program starts from here

optHash=AtlasSNP_SAM.processArguments()
start=AtlasSNP_SAM.new(optHash)
start.hashReference()
start.SNP_list()
start.evaluate()

exit(0);
